/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2006  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "pch.h"

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "ForwardingConfigWindow.h"
#include "Application.h"
#include "Forwarding.h"
#include "ForwardingConfigFile.h"
#include "OperationLog.h"
#include "Log.h"
#include "LogWidget.h"
#include "GlobalState.h"
//#include "Conf.h"
#include "Verify.h"
#include <ace/config-lite.h>
#include <wx/string.h>
#include <wx/frame.h>
#include <wx/event.h>
#include <wx/icon.h>
#include <wx/sizer.h>
#include <wx/gbsizer.h>
#include <wx/panel.h>
#include <wx/stattext.h>
#include <wx/listctrl.h>
#include <wx/button.h>
#include <wx/strconv.h>
#include <wx/msgdlg.h>
#include <wx/scrolwin.h>
#include <wx/bmpbuttn.h>
#include <wx/bitmap.h>
#include <wx/filename.h>
#include <wx/cursor.h>
#include <wx/gdicmn.h>
#include <wx/statusbr.h>
#include <wx/statbox.h>
#include <iterator>
#include <sstream>

namespace wxGUI
{

static ForwardingConfigFile& FCF()
{
	return wxGetApp().forwardingConfigFile();
}

static wxString utf8ToWxString(Forwarding::Utf8String const& utf8)
{
	return wxString(wxConvUTF8.cMB2WC(utf8.raw().c_str()), *wxConvCurrent);
}

static Forwarding::Utf8String wxStringToUtf8(wxString const& wxs)
{
	std::string const utf8(wxConvUTF8.cWC2MB(wxs.wc_str(*wxConvCurrent)));
	return Forwarding::Utf8String(utf8);
}


class ForwardingConfigWindow::Impl : public wxFrame, private AbstractLogView
{
	DECLARE_NON_COPYABLE(Impl)
public:
	Impl();
	
	virtual ~Impl();
	
	static void show();
	
	Forwarding::Config& forwardingConfig() { return m_forwardingConfig; }

	/**
	 * \brief Swaps two options in forwarding config.
	 */
	void swapOptions(long idx1, long idx2);

	void switchToOption(Forwarding::Option const& option);
private:
	bool prepareForWindowDestruction();
	
	bool isModified() const;
	
	void onWindowClose(wxCloseEvent& evt);
	
	void onWindowActivate(wxActivateEvent& evt);
	
	void onSave(wxCommandEvent& evt);

	static Impl* m_spInstance;
	Forwarding::Config m_forwardingConfig;
	OptionsPanel* m_pOptionsPanel;
	ProxyChainPanel* m_pProxyChainPanel;
	
	DECLARE_EVENT_TABLE()
};


class ForwardingConfigWindow::MouseOverEvtHandler : public wxEvtHandler
{
public:
	MouseOverEvtHandler(wxStatusBar* sb, wxString text);

	virtual ~MouseOverEvtHandler();
private:
	void onMouseOver(wxMouseEvent& evt);

	void onMouseOut(wxMouseEvent& evt);

	wxStatusBar* m_pStatusBar;
	wxString m_text;

	DECLARE_EVENT_TABLE()
};


class ForwardingConfigWindow::OptionsPanel : public wxPanel
{
	DECLARE_NON_COPYABLE(OptionsPanel)
public:
	OptionsPanel(wxWindow* parent, Impl& owner);
	
	virtual ~OptionsPanel();
	
	void reload(Forwarding::Config const& config);
private:
	/**
	 * ITEM_REALIZED means that this item represents an option in
	 * m_rOwner.forwardingConfig().options()
	 */
	enum ItemState { ITEM_NOT_REALIZED = 0, ITEM_REALIZED };
	
	void appendOption(Forwarding::Option const& option);
	
	long getSelectedItem() const;

	void setSelectedItem(long item);

	void onNew(wxCommandEvent& evt);
	
	void onRename(wxCommandEvent& evt);
	
	void onDelete(wxCommandEvent& evt);
	
	void onMoveUp(wxCommandEvent& evt);
	
	void onMoveDown(wxCommandEvent& evt);

	void onLabelEditEnd(wxListEvent& evt);
	
	void onItemSelected(wxListEvent& evt);

	Impl& m_rOwner;
	wxListCtrl* m_pOptionsCtrl;
	wxButton* m_pNewBtn;
	wxButton* m_pRenameBtn;
	wxButton* m_pDeleteBtn;
	wxButton* m_pUpBtn;
	wxButton* m_pDownBtn;

	DECLARE_EVENT_TABLE()
};


class ForwardingConfigWindow::ProxyChainPanel : public wxScrolledWindow
{
	DECLARE_NON_COPYABLE(ProxyChainPanel)
public:
	ProxyChainPanel(wxWindow* parent, Impl& owner);
	
	virtual ~ProxyChainPanel();

	void loadOption(Forwarding::Option const& option);

	void clearChain();
private:
	void loadBitmaps();
	
	void addDesktopButton(int id, int col);

	void addForwardButton(int id, int col, wxCursor const& cursor);
	
	void addProxyButton(int id, int col, wxCursor const& cursor);
	
	void addInternetButton(int id, int col);

	void setPreferredSize();

	void onButtonPressed(wxCommandEvent& evt);

	void onResize(wxSizeEvent& evt);

	Impl& m_rOwner;
	wxGridBagSizer* m_pSizer;
	wxBitmap m_desktopBitmap;
	wxBitmap m_proxyBitmap;
	wxBitmap m_proxyGhostBitmap;
	wxBitmap m_internetBitmap;
	wxBitmap m_forwardBitmap;

	DECLARE_EVENT_TABLE()
};


enum {
	ID_LOG_CLEAR_BTN,
	ID_OPTIONS_LIST,
	ID_BTN_NEW,
	ID_BTN_RENAME,
	ID_BTN_DELETE,
	ID_BTN_UP,
	ID_BTN_DOWN,
	ID_CHAIN_BUTTON_FIRST = wxID_HIGHEST + 1 // must be the last one
};

BEGIN_EVENT_TABLE(ForwardingConfigWindow::Impl, wxFrame)
	EVT_ACTIVATE(onWindowActivate)
	EVT_CLOSE(onWindowClose)
	EVT_BUTTON(wxID_SAVE, onSave)
END_EVENT_TABLE()

BEGIN_EVENT_TABLE(ForwardingConfigWindow::MouseOverEvtHandler, wxEvtHandler)
	EVT_ENTER_WINDOW(onMouseOver)
	EVT_LEAVE_WINDOW(onMouseOut)
END_EVENT_TABLE()

BEGIN_EVENT_TABLE(ForwardingConfigWindow::OptionsPanel, wxPanel)
	EVT_BUTTON(ID_BTN_NEW, onNew)
	EVT_BUTTON(ID_BTN_RENAME, onRename)
	EVT_BUTTON(ID_BTN_DELETE, onDelete)
	EVT_BUTTON(ID_BTN_UP, onMoveUp)
	EVT_BUTTON(ID_BTN_DOWN, onMoveDown)
	EVT_LIST_END_LABEL_EDIT(ID_OPTIONS_LIST, onLabelEditEnd)
	EVT_LIST_ITEM_SELECTED(ID_OPTIONS_LIST, onItemSelected)
END_EVENT_TABLE()

BEGIN_EVENT_TABLE(ForwardingConfigWindow::ProxyChainPanel, wxPanel)
	EVT_BUTTON(wxID_ANY, onButtonPressed)
END_EVENT_TABLE()


ForwardingConfigWindow::ForwardingConfigWindow()
:	m_ptrImpl(new Impl)
{
}

ForwardingConfigWindow::~ForwardingConfigWindow()
{
}

void
ForwardingConfigWindow::show()
{
	Impl::show();
}


/*===================== ForwardingConfigWindow::Impl =====================*/

ForwardingConfigWindow::Impl* ForwardingConfigWindow::Impl::m_spInstance = 0;

ForwardingConfigWindow::Impl::Impl()
:	wxFrame(wxGetApp().GetTopWindow(), -1, _T("Forwarding Configuration")),
	AbstractLogView(*OperationLog::instance()),
	m_forwardingConfig(FCF().getForwardingConfig()),
	m_pOptionsPanel(0),
	m_pProxyChainPanel(0)
{
#if 0
	if (m_forwardingConfig.options().empty()) {
		// The idea is to always have at least one option selected,
		// so we must not let the list of options to be empty.
		// Note: this restriction only applies to ForwardingConfigWindow.
		// The rest of BFilter handles the absence of forwarding options
		// just fine.
		ObsoleteForwardingInfo fwd_info(
			GlobalState::ReadAccessor()->config().getObsoleteForwardingInfo()
		);
		fwd_info.toNewFormat().swap(m_forwardingConfig);
		assert(!m_forwardingConfig.options().empty());
	}
#endif
	
	SetExtraStyle(wxWS_EX_BLOCK_EVENTS);
	SetIcon(wxIcon(_T("AppIcon"), wxBITMAP_TYPE_ICO_RESOURCE, 16, 16));
	SetStatusBar(new wxStatusBar(this, -1));

	wxBoxSizer* topsizer = new wxBoxSizer(wxVERTICAL);
	SetSizer(topsizer);
	
	wxPanel* top_panel = new wxPanel(this);
	topsizer->Add(top_panel, 1, wxGROW);
	
	wxBoxSizer* panel_sizer = new wxBoxSizer(wxVERTICAL);
	top_panel->SetSizer(panel_sizer);
	
	m_pOptionsPanel = new OptionsPanel(top_panel, *this);
	panel_sizer->Add(m_pOptionsPanel, 0, wxGROW|wxALL, 3);

	m_pProxyChainPanel = new ProxyChainPanel(top_panel, *this);
	panel_sizer->Add(m_pProxyChainPanel, 0, wxGROW|wxTOP|wxLEFT|wxRIGHT, 5);
	
	wxBoxSizer* buttonsizer = new wxBoxSizer(wxHORIZONTAL);
	panel_sizer->Add(buttonsizer, 0, wxGROW);
	
	wxBoxSizer* clearlog_sizer = new wxBoxSizer(wxHORIZONTAL);
	buttonsizer->Add(clearlog_sizer, 1, wxTOP|wxLEFT|wxRIGHT, 5);
	clearlog_sizer->Add(
		new wxButton(top_panel, ID_LOG_CLEAR_BTN,
		_T("Clear Log")), 0, wxALIGN_LEFT
	);
	
	wxButton* reload_btn = new wxButton(top_panel, -1, _T("Reload"));
	buttonsizer->Add(reload_btn, 0, wxTOP|wxLEFT|wxRIGHT, 5);
	reload_btn->Disable();
	wxButton* save_btn = new wxButton(top_panel, wxID_SAVE, _T("Save"));
	buttonsizer->Add(save_btn, 0, wxTOP|wxLEFT|wxRIGHT, 5);
	
	buttonsizer->Add(1, 1, 1);
	
	wxStaticBoxSizer* logsizer = new wxStaticBoxSizer(
		new wxStaticBox(top_panel, -1, _T("Log")), wxVERTICAL
	);
	panel_sizer->Add(logsizer, 0, wxEXPAND);
	LogWidget* log_widget = new LogWidget(
		OperationLog::instance(),
		top_panel, -1, wxEmptyString,
		wxDefaultPosition, wxSize(-1, 85)
	);
	logsizer->Add(log_widget, 1, wxEXPAND);
	
	m_pOptionsPanel->reload(m_forwardingConfig);

	panel_sizer->SetSizeHints(top_panel);
	topsizer->SetSizeHints(this);
}

ForwardingConfigWindow::Impl::~Impl()
{
}

void
ForwardingConfigWindow::Impl::show()
{
	if (!m_spInstance) {
		m_spInstance = new Impl;
		m_spInstance->Show();
	} else {
		m_spInstance->Show();
		m_spInstance->Raise();
	}
}

void
ForwardingConfigWindow::Impl::onWindowActivate(wxActivateEvent& evt)
{
	AbstractLogView::reportVisibility(evt.GetActive());
}

bool
ForwardingConfigWindow::Impl::prepareForWindowDestruction()
{
	if (!isModified()) {
		return true;
	}
	
	int res = wxMessageBox(
		_T("Are you sure you want to close this window?\n")
		_T("Any unsaved changes will be lost."),
		_T("Question"),
		wxYES_NO, this
	);
	if (res != wxYES) {
		return false;
	}
	return true;
}

bool
ForwardingConfigWindow::Impl::isModified() const
{
	std::ostringstream strm1;
	m_forwardingConfig.toStream(strm1);

	std::ostringstream strm2;
	FCF().getForwardingConfig().toStream(strm2);
	return strm1.str() != strm2.str();
}

void
ForwardingConfigWindow::Impl::onWindowClose(wxCloseEvent& evt)
{
	if (evt.CanVeto()) {
		if (!prepareForWindowDestruction()) {
			evt.Veto();
			return;
		}
	}
	
	if (m_spInstance == this) {
		m_spInstance = 0;
	}
	Destroy();
}

void
ForwardingConfigWindow::Impl::onSave(wxCommandEvent& evt)
{
	FCF().applyAndSave(m_forwardingConfig);
}

void
ForwardingConfigWindow::Impl::swapOptions(long const idx1, long const idx2)
{
	Forwarding::OptionList::iterator it1(m_forwardingConfig.options().begin());
	Forwarding::OptionList::iterator it2(it1);
	std::advance(it1, idx1);
	std::advance(it2, idx2);
	it1->swap(*it2);
}

void
ForwardingConfigWindow::Impl::switchToOption(Forwarding::Option const& opt)
{
	m_pProxyChainPanel->loadOption(opt);
}


/*============== ForwardingConfigWindow::MouseOverEvtHandler =============*/

ForwardingConfigWindow::MouseOverEvtHandler::MouseOverEvtHandler(
	wxStatusBar* sb, wxString text)
:	m_pStatusBar(sb),
	m_text(text)
{
	assert(m_pStatusBar);
}

ForwardingConfigWindow::MouseOverEvtHandler::~MouseOverEvtHandler()
{
}

void
ForwardingConfigWindow::MouseOverEvtHandler::onMouseOver(wxMouseEvent& evt)
{
	evt.Skip();
	m_pStatusBar->SetStatusText(m_text);
}

void
ForwardingConfigWindow::MouseOverEvtHandler::onMouseOut(wxMouseEvent& evt)
{
	evt.Skip();
	
	// When moving cursor from one control to another, it's possible
	// to get mouse-out and mouse-over events in any order.
	if (m_pStatusBar->GetStatusText() == m_text) {
		m_pStatusBar->SetStatusText(wxEmptyString);
	}
}


/*================= ForwardingConfigWindow::OptionsPanel =================*/

ForwardingConfigWindow::OptionsPanel::OptionsPanel(
	wxWindow* parent, Impl& owner)
:	wxPanel(parent),
	m_rOwner(owner),
	m_pOptionsCtrl(0),
	m_pNewBtn(0),
	m_pRenameBtn(0),
	m_pDeleteBtn(0),
	m_pUpBtn(0),
	m_pDownBtn(0)
{
	wxBoxSizer* topsizer = new wxBoxSizer(wxHORIZONTAL);
	SetSizer(topsizer);
	
	{
		m_pOptionsCtrl = new wxListCtrl(
			this, ID_OPTIONS_LIST, wxDefaultPosition, wxSize(140, 100),
			wxLC_REPORT|wxLC_EDIT_LABELS|wxLC_NO_HEADER|wxLC_SINGLE_SEL
		);
		topsizer->Add(m_pOptionsCtrl, 0, wxGROW);
		int width = 0;
		int height = 0;
		m_pOptionsCtrl->GetClientSize(&width, &height);
		wxListItem header;
		header.SetWidth(width);
		m_pOptionsCtrl->InsertColumn(0, header);
		m_pOptionsCtrl->PushEventHandler(
			new MouseOverEvtHandler(
				m_rOwner.GetStatusBar(), _T("Use context menu")
			)
		);
	}
	
	
	wxBoxSizer* btn_sizer = new wxBoxSizer(wxVERTICAL);
	topsizer->Add(btn_sizer, 0, wxLEFT, 6);
	
	m_pNewBtn = new wxButton(this, ID_BTN_NEW, _T("New"));
	btn_sizer->Add(m_pNewBtn, 0, wxGROW|wxLEFT|wxRIGHT|wxBOTTOM, 3);
	
	m_pRenameBtn = new wxButton(this, ID_BTN_RENAME, _T("Rename"));
	btn_sizer->Add(m_pRenameBtn, 0, wxGROW|wxALL, 3);
	
	m_pDeleteBtn = new wxButton(this, ID_BTN_DELETE, _T("Delete"));
	btn_sizer->Add(m_pDeleteBtn, 0, wxGROW|wxALL, 3);
	
	m_pUpBtn = new wxButton(this, ID_BTN_UP, _T("Up"));
	btn_sizer->Add(m_pUpBtn, 0, wxGROW|wxALL, 3);
	
	m_pDownBtn = new wxButton(this, ID_BTN_DOWN, _T("Down"));
	btn_sizer->Add(m_pDownBtn, 0, wxGROW|wxLEFT|wxRIGHT|wxTOP, 3);
	
	topsizer->SetSizeHints(this);
}

ForwardingConfigWindow::OptionsPanel::~OptionsPanel()
{
}

void
ForwardingConfigWindow::OptionsPanel::reload(Forwarding::Config const& config)
{
	m_pOptionsCtrl->DeleteAllItems();

	Forwarding::OptionList const& options = config.options();
	Forwarding::OptionList::const_iterator it(options.begin());
	Forwarding::OptionList::const_iterator const end(options.end());
	for (; it != end; ++it) {
		appendOption(*it);
	}
	if (getSelectedItem() == -1) {
		setSelectedItem(0);
	}
}

void
ForwardingConfigWindow::OptionsPanel::appendOption(
	Forwarding::Option const& option)
{
	long const row = m_pOptionsCtrl->GetItemCount();
	m_pOptionsCtrl->InsertItem(row, utf8ToWxString(option.getName()));
	m_pOptionsCtrl->SetItemData(row, ITEM_REALIZED);
	if (option.isSelected()) {
		setSelectedItem(row);
	}
}

/**
 * \return selected item index or -1 if no item is selected.
 */
long
ForwardingConfigWindow::OptionsPanel::getSelectedItem() const
{
	return m_pOptionsCtrl->GetNextItem(
		-1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED
	);
}

void
ForwardingConfigWindow::OptionsPanel::setSelectedItem(long item)
{
	m_pOptionsCtrl->SetItemState(
		item,
		wxLIST_STATE_FOCUSED|wxLIST_STATE_SELECTED,
		wxLIST_STATE_FOCUSED|wxLIST_STATE_SELECTED
	);
}

void
ForwardingConfigWindow::OptionsPanel::onNew(wxCommandEvent& evt)
{
	long const row = m_pOptionsCtrl->GetItemCount();
	m_pOptionsCtrl->InsertItem(row, wxEmptyString);
	m_pOptionsCtrl->SetItemData(row, ITEM_NOT_REALIZED);
	m_pOptionsCtrl->EditLabel(row);
}

void
ForwardingConfigWindow::OptionsPanel::onRename(wxCommandEvent& evt)
{
	long const row = getSelectedItem();
	if (row >= 0) {
		m_pOptionsCtrl->EditLabel(row);
	}
}

void
ForwardingConfigWindow::OptionsPanel::onDelete(wxCommandEvent& evt)
{
	long const row = getSelectedItem();
	if (row < 0) {
		return;
	}

	int const res = wxMessageBox(
		_T("Really delete this profile?"),
		_T("Delete Confirmation"),
		wxYES_NO, this
	);
	if (res != wxYES) {
		return;
	}
	
	if (m_pOptionsCtrl->GetItemData(row) == ITEM_REALIZED) {
		Forwarding::Config& config = m_rOwner.forwardingConfig();
		Forwarding::OptionList::iterator it(config.options().begin());
		std::advance(it, row);
		config.options().erase(it);
	}
	m_pOptionsCtrl->DeleteItem(row);
}

void
ForwardingConfigWindow::OptionsPanel::onMoveUp(wxCommandEvent& evt)
{
	long const item = getSelectedItem();
	long const prev = item - 1;
	if (prev >= 0 /*&& item < m_pOptionsCtrl->GetItemCount()*/) {
		wxString const text(m_pOptionsCtrl->GetItemText(prev));
		long const data(m_pOptionsCtrl->GetItemData(prev));
		m_pOptionsCtrl->InsertItem(item + 1, text);
		m_pOptionsCtrl->SetItemData(item + 1, data);
		m_pOptionsCtrl->DeleteItem(prev);
		m_rOwner.swapOptions(item, prev);
	}
}

void
ForwardingConfigWindow::OptionsPanel::onMoveDown(wxCommandEvent& evt)
{
	long const item = getSelectedItem();
	long next = item + 1;
	if (item >= 0 && next < m_pOptionsCtrl->GetItemCount()) {
		wxString const text(m_pOptionsCtrl->GetItemText(next));
		long const data(m_pOptionsCtrl->GetItemData(next));
		m_pOptionsCtrl->InsertItem(item, text); // moves item and next downwards
		m_pOptionsCtrl->SetItemData(item, data);
		m_pOptionsCtrl->DeleteItem(next + 1);
		m_rOwner.swapOptions(item, next);
	}
}

void
ForwardingConfigWindow::OptionsPanel::onLabelEditEnd(wxListEvent& evt)
{
	long const row = evt.GetIndex();
	assert(row >= 0);
	
	bool const renaming = (m_pOptionsCtrl->GetItemData(row) == ITEM_REALIZED);

	if (evt.IsEditCancelled() && !renaming) {
		m_pOptionsCtrl->DeleteItem(row);
		return;
	}

	if (evt.GetLabel().empty()) {
		evt.Veto();
		return;
	}
	
	Forwarding::Utf8String option_name(wxStringToUtf8(evt.GetLabel()));
	Forwarding::Config& config = m_rOwner.forwardingConfig();
	Forwarding::OptionList::iterator it(config.options().begin());
	std::advance(it, row);
	
	if (renaming) {
		it->setName(option_name);
	} else { // creating
		Forwarding::OptionList::iterator ins_pos(
			config.options().insert(it, Forwarding::Option(option_name))
		);
		m_pOptionsCtrl->SetItemData(row, ITEM_REALIZED);
		m_rOwner.switchToOption(*ins_pos);
	}
}

void
ForwardingConfigWindow::OptionsPanel::onItemSelected(wxListEvent& evt)
{
	long const row = evt.GetIndex();
	assert(row >= 0);
	
	if (m_pOptionsCtrl->GetItemData(row) == ITEM_NOT_REALIZED) {
		return;
	}

	Forwarding::Config& config = m_rOwner.forwardingConfig();
	Forwarding::OptionList::iterator it(config.options().begin());
	std::advance(it, row);

	m_rOwner.switchToOption(*it);
}


/*================ ForwardingConfigWindow::ProxyChainPanel ===============*/

ForwardingConfigWindow::ProxyChainPanel::ProxyChainPanel(
	wxWindow* parent, Impl& owner)
:	wxScrolledWindow(parent, -1),
	m_rOwner(owner),
	m_pSizer(0)
{
	loadBitmaps();

	SetScrollRate(32, 0);

	setPreferredSize();
}

ForwardingConfigWindow::ProxyChainPanel::~ProxyChainPanel()
{
}

void
ForwardingConfigWindow::ProxyChainPanel::loadBitmaps()
{
	wxFileName fname(wxGetApp().getResourcesDir());

	fname.SetName(_T("desktop.png"));
	m_desktopBitmap.LoadFile(fname.GetFullPath(), wxBITMAP_TYPE_PNG);
	
	fname.SetName(_T("proxy.png"));
	m_proxyBitmap.LoadFile(fname.GetFullPath(), wxBITMAP_TYPE_PNG);
	
	fname.SetName(_T("proxy-ghost.png"));
	m_proxyGhostBitmap.LoadFile(fname.GetFullPath(), wxBITMAP_TYPE_PNG);

	fname.SetName(_T("internet.png"));
	m_internetBitmap.LoadFile(fname.GetFullPath(), wxBITMAP_TYPE_PNG);

	fname.SetName(_T("forward.png"));
	m_forwardBitmap.LoadFile(fname.GetFullPath(), wxBITMAP_TYPE_PNG);
}

void
ForwardingConfigWindow::ProxyChainPanel::setPreferredSize()
{
	// create a dummy option with two proxies
	Forwarding::Utf8String const option_name;
	Forwarding::Option dummy_option(option_name);
	for (int i = 0; i < 2; ++i) {
		dummy_option.proxyChain().push_back(ProxyDescriptor());
	}
	
	loadOption(dummy_option);
	
	// get the size to fit a chain with two proxies
	int client_w = 0, client_h = 0, tmp = 0;
	GetVirtualSize(&client_w, &client_h);
	
	// force horizontal scroll bar to appear
	SetClientSize(client_w / 2, client_h);
	int min_h = 0;
	GetSize(&tmp, &min_h);
	
	// force scroll bar to disappear
	SetClientSize(client_w, client_h);
	int min_w = 0;
	GetSize(&min_w, &tmp);

	SetMinSize(wxSize(min_w, min_h)); 

	clearChain();
}

void
ForwardingConfigWindow::ProxyChainPanel::clearChain()
{
	SetSizer(0);
	m_pSizer = 0;
	DestroyChildren();
}

void
ForwardingConfigWindow::ProxyChainPanel::loadOption(Forwarding::Option const& opt)
{
	wxCursor const hand_cursor(wxCURSOR_HAND);
	
	wxBoxSizer* topsizer = new wxBoxSizer(wxVERTICAL);
	SetSizer(topsizer);
	DestroyChildren();

	m_pSizer = new wxGridBagSizer(/* vgap = */3, /* hgap = */0);
	topsizer->Add(m_pSizer, 0, wxALIGN_CENTER);
	m_pSizer->SetFlexibleDirection(wxBOTH);
	
	int id = ID_CHAIN_BUTTON_FIRST;
	
	addDesktopButton(id, 0);	
	addForwardButton(++id, 1, hand_cursor); // takes 2 cols
	
	m_pSizer->Add(
		new wxStaticText(this, -1, _T("This Computer")),
		wxGBPosition(1, 0), wxGBSpan(1, 2),
		wxALIGN_LEFT|wxALIGN_CENTER_VERTICAL
	);
	
	int col = 3;
	
	typedef Forwarding::ProxyDescriptorList::const_iterator Iter;
	Iter it(opt.proxyChain().begin());
	Iter const end(opt.proxyChain().end());
	for (; it != end; ++it, col += 3) {
		ProxyDescriptor const& proxy = *it;
		
		addProxyButton(++id, col, hand_cursor);
		addForwardButton(++id, col + 1, hand_cursor); // takes 2 cols

		m_pSizer->Add(
			new wxStaticText(this, -1, proxy.getAddr().toString().c_str()),
			wxGBPosition(1, col - 1), wxGBSpan(1, 3),
			wxALIGN_CENTER_HORIZONTAL|wxALIGN_CENTER_VERTICAL
		);
	}

	addInternetButton(++id, col);

	m_pSizer->Add(
		new wxStaticText(this, -1, _T("Internet")),
		wxGBPosition(1, col), wxGBSpan(1, 1),
		wxALIGN_CENTER_HORIZONTAL|wxALIGN_CENTER_VERTICAL
	);
	
	topsizer->FitInside(this);
}

void
ForwardingConfigWindow::ProxyChainPanel::addDesktopButton(int id, int col)
{
	wxBitmapButton* desktop_btn = new wxBitmapButton(
		this, id, m_desktopBitmap, wxDefaultPosition, wxDefaultSize, 0
	);
	m_pSizer->Add(
		desktop_btn, wxGBPosition(0, 0), wxDefaultSpan,
		wxALIGN_CENTER_HORIZONTAL|wxALIGN_CENTER_VERTICAL
	);
	desktop_btn->Disable();
}

void
ForwardingConfigWindow::ProxyChainPanel::addForwardButton(
	int id, int col, wxCursor const& cursor)
{
	wxBitmapButton* forward_btn = new wxBitmapButton(
		this, id, m_forwardBitmap, wxDefaultPosition, wxDefaultSize, 0
	);
	m_pSizer->Add(
		forward_btn, wxGBPosition(0, col), wxGBSpan(1, 2),
		wxALIGN_CENTER_HORIZONTAL|wxALIGN_CENTER_VERTICAL
	);
	forward_btn->SetBitmapHover(m_proxyGhostBitmap);
	forward_btn->SetCursor(cursor);
	forward_btn->PushEventHandler(
		new MouseOverEvtHandler(
			m_rOwner.GetStatusBar(), _T("Add proxy")
		)
	);
}

void
ForwardingConfigWindow::ProxyChainPanel::addProxyButton(
	int id, int col, wxCursor const& cursor)
{
	wxBitmapButton* proxy_btn = new wxBitmapButton(
		this, id, m_proxyBitmap, wxDefaultPosition, wxDefaultSize, 0
	);
	m_pSizer->Add(
		proxy_btn, wxGBPosition(0, col), wxDefaultSpan,
		wxALIGN_CENTER_HORIZONTAL|wxALIGN_CENTER_VERTICAL
	);
	proxy_btn->SetCursor(cursor);
	proxy_btn->PushEventHandler(
		new MouseOverEvtHandler(
			m_rOwner.GetStatusBar(), _T("Edit proxy")
		)
	);
}

void
ForwardingConfigWindow::ProxyChainPanel::addInternetButton(int id, int col)
{
	wxBitmapButton* internet_btn = new wxBitmapButton(
		this, id, m_internetBitmap, wxDefaultPosition, wxDefaultSize, 0
	);
	m_pSizer->Add(
		internet_btn, wxGBPosition(0, col), wxDefaultSpan,
		wxALIGN_LEFT|wxALIGN_CENTER_VERTICAL
	);
	internet_btn->Disable();
}

void
ForwardingConfigWindow::ProxyChainPanel::onButtonPressed(wxCommandEvent& evt)
{
	Refresh(); // workaround a bug in wxWidgets

	int const button_idx = evt.GetId() - ID_CHAIN_BUTTON_FIRST;
	assert(button_idx >= 0);
	
	/*
	Buttons are indexed like this:
	------------------------------
	[0] => This Comuter
	[1] => Arrow
	for (int i = 0; i < NUM_PROXIES; ++i) {
		[i*2]   => Proxy
		[i*2+1] => Arrow
	}
	[2+NUM_PROXIES*2] => Internet
	------------------------------
	The first one and the last one are unclickable, so this leaves us with
	Proxy buttons at even positions and Arrow buttons at odd positions.
	*/
	
	if (button_idx & 1) {
		// arrow
		int const next_proxy_idx = (button_idx - 1) / 2;
		int const prev_proxy_idx = next_proxy_idx - 1;
	} else {
		// proxy
		int const proxy_idx = button_idx / 2 - 1;
	}

	
}

} // namespace wxGUI
